#!/usr/bin/env perl
#
use strict;
use Data::Dumper;
use Getopt::Long;

my $validate_only = 0;
GetOptions("validate-only+" => \$validate_only);

my %parse_func =
  ("O:" => \&parse_object,
   "a:" => \&parse_array,
   "s:" => \&parse_string,
   "i:" => \&parse_integer,
   "b:" => \&parse_integer,
   "N;" => \&parse_null);

# Slurp the stdin
undef $/;
my $data = <>;

my ($tree, $index) = &parse(0);
print "$index bytes parsed";

if ($validate_only) {
  print " -- valid\n";
} else {
  print "\n";
  $Data::Dumper::Indent = 1;
  print Dumper($tree);
}

sub parse {
  my $index = shift;

  my $type = substr($data, $index, 2);
  my $func = $parse_func{$type};
  if (!$func) {
    die "Unknown type '$type' at position $index";
  }
  &$func($index+2);
}

sub parse_object {
  my $index = shift;

  my $self = {};

  my ($name, $members);
  ($name, $index) = &get_string($index);
  $index = skip($index, ':');
  ($members, $index) = &parse_array($index);

  $self->{"$name (class)"} = $members;

  return ($self, $index);
}

sub parse_null {
  my $index = shift;

  return ("<null>", $index);
}

sub parse_integer {
  my $index = shift;

  my $value;
  my $self = {};
  ($value, $index) = get_number($index);
  $index = skip($index, ';');

  return ($value, $index);
}

sub parse_string {
  my $index = shift;

  my $value;
  my $size = 0;

  ($value, $index) = get_string($index);
  $index = skip($index, ';');

  return ($value, $index);
}

sub parse_array {
  my $index = shift;

  my $self = [];
  my $count = 0;
  ($count, $index) = get_number($index);
  $index = skip($index, ':{');
  for (my $i = 0; $i < $count; $i++) {
    my ($key, $value);
    ($key, $index) = &parse($index);
    ($value, $index) = &parse($index);
    push(@$self, { $key => $value });
  }
  $index = skip($index, '}');
  return ($self, $index);
}

sub get_string {
  my $index = shift;

  my $size;
  my $value;
  ($size, $index) = get_number($index);
  $index = skip($index, ':"');
  $value = substr($data, $index, $size);
  $index += $size;
  $index = skip($index, '"');

  return ($value, $index);
}

sub get_number {
  my $index = shift;

  my $ch;
  my $num = undef;
  while(($ch = substr($data, $index, 1)) =~ /\d/) {
    $index++;
    if (!defined($num)) {
      $num = 0;
    }
    $num = 10 * $num + int($ch);
  }

  if (!defined($num)) {
    die "Expected number at $index, found '$ch' instead\n";
  }

  return ($num, $index);
}


sub skip {
  my $index = shift;
  my $expected = shift;

  my $size = length($expected);
  my $actual = substr($data, $index, $size);
  if ($actual ne $expected) {
    die "Expected '$expected' at $index, found '$actual' instead\n";
    exit;
  }

  return $index + $size;
}

__END__
while(<>) {
  foreach $line (split(/([{\}])/)) {
    $indent -= ($line eq "}");
    next unless $line;
    $line =~ s/;/";\n" . "  " x $indent/ge;
    print "  " x $indent, "$line\n";
    $indent += ($line eq "{");
  }
}
